use crate::js_net::{set_init_state, start_network_services};
use crate::mgr_pid::mgr_pid_handle;
use crate::{
    enable_hotfix, get_gray_last_v, init_console, init_default_gray, init_http_listener_pid,
    init_listener_pid, init_pt, reigster_vms_events, CONSOLE_SHELL, FILES_ASYNC_RUNTIME,
    LOAD_PROJECT_SENDER, MAIN_CONDVAR, MAIN_RUN_STATUS, PROJECT_GRAY,
};
use clap::ArgMatches;
use futures::Future;
use num_cpus::get_physical;
use parking_lot::Mutex;
use pi_async::rt::serial_local_thread::LocalTaskRuntime;
use pi_async::rt::{spawn_worker_thread, AsyncRuntimeExt};
use pi_async_file::file::{AsyncFile, AsyncFileOptions};
use pi_core::console::set_console_shell_ctrlc_handler;
use pi_core::terminal::VmTerminal;
use pi_core::{create_snapshot_vm, finish_snapshot, init_ext_vm, init_snapshot, init_work_vm};
use vm_builtin::{task_pool::VmTaskPool, ContextHandle};
use vm_builtin::{AsyncVmTaskPool, VmStartupSnapshot, VmTiming};
use vm_core::vm::{self, Vm};

use std::env;
use std::fs::read_to_string;
use std::path::{Path, PathBuf};
use std::process::Output;
use std::sync::Arc;

pub async fn read_init_source(init_exec_path: String) -> String {
    let source = match AsyncFile::open(
        FILES_ASYNC_RUNTIME.clone(),
        init_exec_path.clone(),
        AsyncFileOptions::OnlyRead,
    )
    .await
    {
        Ok(file) => match file.read(0, file.get_size() as usize).await {
            Ok(f) => f,
            Err(e) => {
                panic!("read init-file path: {:?}, error: {:?}", &init_exec_path, e);
            }
        },
        Err(e) => {
            panic!(
                "open init-file failed path: {:?}, error: {:?}",
                &init_exec_path, e
            );
        }
    };

    match String::from_utf8(source) {
        Ok(s) => s,
        Err(e) => {
            panic!("init-file is not valid utf8 string, error: {:?}", e);
        }
    }
}

pub async fn init_js(
    is_debug_mode: bool,
    init_vm: vm::Vm<VmTaskPool>,
    handle: ContextHandle,
    init_exec_path: String,
    project: String,
    gray: usize,
) {
    // let init_exec_path = if is_debug_mode {
    //     "../dst_server/pi_pt/debug_init.js".to_string()
    // } else {
    //     matches.value_of("init-file").unwrap().to_string()
    // };

    info!(" init-file path: {:?}", init_exec_path);

    let source = read_init_source(init_exec_path.clone()).await;
    let source = format!(
        r#"var project = "{}";
        var gray = {};
        {}"#,
        project, gray, source
    );
    if let Err(e) = init_vm
        .execute(handle, &init_exec_path, source.as_str())
        .await
    {
        panic!(
            "load init-file failed path: {:?}, error: {:?}",
            init_exec_path, e
        );
    }
}

pub fn init_main_sync(
    init_heap_size: Option<usize>,
    max_heap_size: Option<usize>,
    debug_port: Option<u16>,
    debug_break: bool,
    work_vm_count: usize,
    init_exec_path: String,
    is_console: bool,
    project: String,
    ext_vm_count: usize,
) -> bool {
    if let Some(sender) = LOAD_PROJECT_SENDER.lock().as_ref() {
        match sender.send((
            init_heap_size,
            max_heap_size,
            debug_port,
            debug_break,
            work_vm_count,
            init_exec_path,
            is_console,
            project,
            ext_vm_count,
        )) {
            Ok(_) => true,
            Err(_) => false,
        }
    } else {
        false
    }
}

pub async fn init_main(
    init_heap_size: Option<usize>,
    max_heap_size: Option<usize>,
    debug_port: Option<u16>,
    debug_break: bool,
    work_vm_count: usize,
    init_exec_path: String,
    is_console: bool,
    project: String,
    ext_vm_count: usize,
    new_work_rts: Arc<Mutex<Vec<LocalTaskRuntime>>>,
) {
    let pool = VmTaskPool::default();
    let timer = pool.to_timer();
    let pool_timer = pool.to_timer();
    let mut init_vm = create_snapshot_vm(
        pool,
        pool_timer,
        init_heap_size,
        max_heap_size,
        debug_port,
        MAIN_RUN_STATUS.clone(),
        MAIN_CONDVAR.clone(),
    );
    let init_vm_runner = init_vm.take_runner().unwrap();
    let queue_len_getter = init_vm_runner.get_inner_handler();

    //启动初始虚拟机线程，并运行初始虚拟机
    let init_vm_handle = spawn_worker_thread(
        "Init-Vm",
        2 * 1024 * 1024,
        MAIN_RUN_STATUS.clone(),
        MAIN_CONDVAR.clone(),
        10,
        Some(10),
        move || {
            let now = minstant::Instant::now();
            timer.tick();
            let len = init_vm_runner.run_once();
            (len == 0, minstant::Instant::now().duration_since(now))
        },
        move || {
            if let Some(len) = queue_len_getter.len() {
                len
            } else {
                0
            }
        },
    );

    let init_vm = init_vm.init().unwrap();

    let snapshot_context = init_snapshot(&init_vm).await;
    let gray = get_gray_last_v();
    init_js(
        debug_port.is_some(),
        init_vm.clone(),
        snapshot_context.clone(),
        init_exec_path,
        project.clone(),
        gray,
    )
    .await;
    finish_snapshot(&init_vm, snapshot_context).await;

    let mut worker_pool_timers = Vec::with_capacity(work_vm_count);
    for _ in 0..work_vm_count {
        let pool = VmTaskPool::default();
        let timer = pool.to_timer();
        worker_pool_timers.push((pool, timer));
    }
    let snapshot = VmStartupSnapshot::Snapshot(init_vm.get_snapshot().await.unwrap());
    let snapshot_bytes = snapshot.bytes();
    let workers = init_work_vm(
        snapshot_bytes.clone(),
        worker_pool_timers,
        init_heap_size,
        max_heap_size,
        debug_port,
        debug_break,
        "PI-SERV",
        8,
        10,
        None,
        move |work_vm_timer: VmTaskPool, work_vm_runner: vm::VmRunner<VmTaskPool>| {
            move || {
                let now = minstant::Instant::now();
                work_vm_timer.tick();
                let len = work_vm_runner.run_once();
                (len == 0, minstant::Instant::now().duration_since(now))
            }
        },
    )
    .await;

    let vms: Vec<vm::Vm<VmTaskPool>> = workers.iter().map(|(_, vm)| vm.clone()).collect();

    reigster_vms_events(&vms, debug_port.is_some());
    // 创建管理pid
    mgr_pid_handle(&vms);
    // 设置灰度对应的虚拟机
    init_default_gray(vms.clone(), gray);
    PROJECT_GRAY.write().insert(project.clone(), gray);

    // 设置扩展虚拟机
    ext_vm(
        snapshot_bytes.clone(),
        init_heap_size,
        max_heap_size,
        debug_port,
        debug_break,
        ext_vm_count,
        gray,
    )
    .await;
    // 初始化平台代码
    init_pt(workers[0].clone().1).await;

    //所有虚拟机启动完成之后创建listener pid
    init_listener_pid(gray);
    init_http_listener_pid(gray);

    let mut recv_buf_size = 131072;
    if let Ok(var) = env::var("RECV_BUF_SIZE") {
        recv_buf_size = var.parse().unwrap_or(recv_buf_size);
    }

    let mut send_buf_size = 131072;
    if let Ok(var) = env::var("SEND_BUF_SIZE") {
        send_buf_size = var.parse().unwrap_or(send_buf_size);
    }

    // 最后启动网络服务
    let _ = start_network_services(
        recv_buf_size,
        send_buf_size,
        16384,
        100000,
        2048,
        2097152,
        1,
        new_work_rts,
    );
    // 重新设置pi_pt状态
    set_init_state(false);
    if is_console {
        if let Some((worker, worker_context)) = init_console(&workers).await {
            if let Some(console_shell) = CONSOLE_SHELL.read().as_ref() {
                //启动控制台
                let vid = worker.get_vid();
                let cid = worker_context.0;
                let console_shell = console_shell.clone();

                //设置控制台Ctrl+C的处理器
                let prompt_prefix = "Pid{".to_string()
                    + vid.to_string().as_str()
                    + ", "
                    + cid.to_string().as_str()
                    + "}";
                set_console_shell_ctrlc_handler(console_shell.clone(), prompt_prefix.clone());

                console_shell.set_prompt_prefix(prompt_prefix.as_str());
                console_shell.show_welcome_info(
                    "PiServ",
                    "0.2.0",
                    Some("Workers: ".to_string() + workers.len().to_string().as_str()),
                );
                console_shell.wait_input(prompt_prefix);
            }
        } else {
            //不启动控制台，且需要分析堆内存，则设置Ctrl+C的默认处理器
            #[cfg(any(
                all(
                    target_arch = "x86_64",
                    target_os = "linux",
                    feature = "jemalloc_profiling"
                ),
                feature = "profiling_heap"
            ))]
            set_default_ctrlc_handler();
        }
    }
    enable_hotfix(gray, project);
}

// 初始化扩展虚拟机用于定时任务、手动创建的pid
async fn ext_vm(
    snapshot_bytes: &[u8],
    init_heap_size: Option<usize>,
    max_heap_size: Option<usize>,
    debug_port: Option<u16>,
    debug_break: bool,
    ext_vm_count: usize,
    gray: usize,
) {
    let mut worker_pool_timers = Vec::with_capacity(ext_vm_count);
    for _ in 0..ext_vm_count {
        let pool = VmTaskPool::default();
        let timer = pool.to_timer();
        worker_pool_timers.push((pool, timer));
    }
    let workers = init_ext_vm(
        snapshot_bytes,
        worker_pool_timers,
        init_heap_size,
        max_heap_size,
        debug_port,
        debug_break,
        "PI-SERV-EXT",
        8,
        10,
        None,
        move |work_vm_timer: VmTaskPool, work_vm_runner: vm::VmRunner<VmTaskPool>| {
            move || {
                let now = minstant::Instant::now();
                work_vm_timer.tick();
                let len = work_vm_runner.run_once();
                (len == 0, minstant::Instant::now().duration_since(now))
            }
        },
    )
    .await;

    let vms: Vec<vm::Vm<VmTaskPool>> = workers.iter().map(|(_, vm)| vm.clone()).collect();

    reigster_vms_events(&vms, debug_port.is_some());
    // 创建管理pid
    mgr_pid_handle(&vms);
    // 设置灰度对应的虚拟机
    init_default_gray(vms.clone(), gray + 1);
}
